from io import BytesIO
from math import inf

from mcap_ros2._cdr import CdrReader, CdrWriter, EncapsulationKind

tf2_msg__TFMessage = (
    "0001000001000000cce0d158f08cf9060a000000626173655f6c696e6b00000006000000"
    "7261646172000000ae47e17a14ae0e400000000000000000000000000000000000000000"
    "0000000000000000000000000000000000000000000000000000f03f"
)
rcl_interfaces__ParameterEvent = (
    "00010000a9b71561a570ea01110000002f5f726f7332636c695f33373833363300000000"
    "010000000d0000007573655f73696d5f74696d6500010000000000000000000000000000"
    "000000000100000000000000000000000000000000000000000000000000000000000000"
    "00000000"
)
std_msgs__Empty = "0001000000"


def test_parse_tfmessage():
    data = bytes.fromhex(tf2_msg__TFMessage)
    reader = CdrReader(data)
    assert reader.byte_length() == len(data)
    assert reader.decoded_bytes() == 4
    assert reader.kind() == EncapsulationKind.CDR_LE

    # geometry_msgs/TransformStamped[] transforms
    assert reader.sequence_length() == 1
    # std_msgs/Header header
    assert reader.uint32() == 1490149580  # uint32 sec
    assert reader.int32() == 117017840  # int32 nanosec
    assert reader.string() == "base_link"  # frame_id
    assert reader.string() == "radar"  # child_frame_id
    # geometry_msgs/Transform transform
    assert reader.float64() == 3.835  # float64 x
    assert reader.float64() == 0.0  # float64 y
    assert reader.float64() == 0.0  # float64 z
    assert reader.float64() == 0.0  # float64 x
    assert reader.float64() == 0.0  # float64 y
    assert reader.float64() == 0.0  # float64 z
    assert reader.float64() == 1.0  # float64 w

    assert reader.offset == len(data)
    assert reader.decoded_bytes() == len(data)


def test_parse_parameter_event():
    data = bytes.fromhex(rcl_interfaces__ParameterEvent)
    reader = CdrReader(data)

    # builtin_interfaces/Time stamp
    assert reader.uint32() == 1628813225  # uint32 sec
    assert reader.int32() == 32141477  # int32 nanosec
    # string node
    assert reader.string() == "/_ros2cli_378363"
    # Parameter[] new_parameters
    assert reader.sequence_length() == 1
    assert reader.string() == "use_sim_time"  # string name
    # ParameterValue value
    assert reader.uint8() == 1  # uint8 type
    assert reader.int8() == 0  # bool bool_value
    assert reader.int64() == 0  # int64 integer_value
    assert reader.float64() == 0  # float64 double_value
    assert reader.string() == ""  # string string_value

    assert reader.sequence_length() == 0
    assert reader.int8_array(0) == []  # byte[] byte_array_value
    assert reader.sequence_length() == 0
    assert reader.uint8_array(0) == bytes()  # bool[] bool_array_value
    assert reader.sequence_length() == 0
    assert reader.int64_array(0) == []  # int64[] integer_array_value
    assert reader.sequence_length() == 0
    assert reader.float64_array(0) == []  # float64[] double_array_value
    assert reader.sequence_length() == 0
    assert reader.string_array(0) == []  # string[] string_array_value

    # Parameter[] changed_parameters
    assert reader.sequence_length() == 0
    # Parameter[] deleted_parameters
    assert reader.sequence_length() == 0

    assert reader.decoded_bytes() == len(data)


def test_parse_empty_msg():
    data = bytes.fromhex(std_msgs__Empty)
    reader = CdrReader(data)
    assert reader.uint8() == 0  # uint8 structure_needs_at_least_one_member
    assert reader.decoded_bytes() == len(data)


def test_read_big_endian():
    data = bytes.fromhex("000100001234000056789abcdef0000000000000")
    reader = CdrReader(data)
    assert reader.uint16BE() == 0x1234
    assert reader.uint32BE() == 0x56789ABC
    assert reader.uint64BE() == 0xDEF0000000000000


def test_seeking():
    data = bytes.fromhex(tf2_msg__TFMessage)
    reader = CdrReader(data)

    reader.seek_to(4 + 4 + 4 + 4 + 4 + 10 + 4 + 6)
    assert reader.float64() == 3.835

    # This works due to aligned reads
    reader.seek_to(4 + 4 + 4 + 4 + 4 + 10 + 4 + 3)
    assert reader.float64() == 3.835

    reader.seek(-8)
    assert reader.float64() == 3.835
    assert reader.float64() == 0.0


def test_arrays():
    INT64_MAX = 9223372036854775807
    INT64_MIN = -9223372036854775808
    UINT64_MAX = 0xFFFFFFFFFFFFFFFF
    FLOAT1 = -123.19999694824219
    FLOAT2 = 123.56999969482422

    for [write_fn, read_fn, array, output] in (
        ["write_boolean_array", "boolean_array", [True, False, True], None],
        ["write_int8_array", "int8_array", [-128, -1, 0, 1, 127], None],
        ["write_uint8_array", "uint8_array", [0, 128, 255], b"\x00\x80\xff"],
        ["write_int16_array", "int16_array", [-32767, -1, 0, 1, 32767], None],
        ["write_uint16_array", "uint16_array", [0, 0xFF, 0xFFFF], None],
        ["write_int32_array", "int32_array", [-2147483648, -1, 0, 1, 2147483647], None],
        ["write_uint32_array", "uint32_array", [0, 0xFF, 0xFFFF, 0xFFFFFFFF], None],
        ["write_int64_array", "int64_array", [INT64_MIN, -1, 0, 1, INT64_MAX], None],
        ["write_uint64_array", "uint64_array", [0, UINT64_MAX], None],
        [
            "write_float32_array",
            "float32_array",
            [FLOAT1, -1.0, 0.0, 1.0, FLOAT2, inf, -inf, 42.0],
            None,
        ],
        [
            "write_float64_array",
            "float64_array",
            [FLOAT1, -1.0, 0.0, 1.0, FLOAT2, inf, -inf, 42.0],
            None,
        ],
        ["write_string_array", "string_array", ["", "a", "ab", "Test!", "↑↓"], None],
    ):
        assert isinstance(write_fn, str)
        assert isinstance(read_fn, str)
        assert isinstance(array, list)

        buffer = BytesIO()
        writer = CdrWriter(buffer)
        writer.write_uint32(len(array))
        getattr(writer, write_fn)(array)

        data = buffer.getvalue()
        reader = CdrReader(data)
        output = output or array
        assert reader.sequence_length() == len(array)
        assert getattr(reader, read_fn)(len(array)) == output
        assert reader.decoded_bytes() == len(data)



from mcap_ros2.decoder import DecoderFactory

from mcap.reader import make_reader

from .generate import generate_sample_data


def test_ros2_decoder():
    with generate_sample_data() as m:
        reader = make_reader(m, decoder_factories=[DecoderFactory()])
        count = 0
        for index, (_, _, _, ros_msg) in enumerate(
            reader.iter_decoded_messages("/chatter")
        ):
            assert ros_msg.data == f"string message {index}"
            assert ros_msg._type == "std_msgs/String"
            assert ros_msg._full_text == "# std_msgs/String\nstring data"
            count += 1
        assert count == 10

        count = 0
        for _, _, _, ros_msg in reader.iter_decoded_messages("/empty"):
            assert ros_msg._type == "std_msgs/Empty"
            assert ros_msg._full_text == "# std_msgs/Empty"
            count += 1
        assert count == 10


def test_ros2_decoder_msg_eq():
    with generate_sample_data() as m:
        reader = make_reader(m, decoder_factories=[DecoderFactory()])

        decoded_messages = reader.iter_decoded_messages("/chatter")
        _, _, _, msg0 = next(decoded_messages)
        _, _, _, msg1 = next(decoded_messages)
        assert msg0.data == "string message 0"
        assert msg1.data == "string message 1"
        assert msg0 == msg0 and msg1 == msg1
        assert msg0 != msg1 and msg1 != msg0



from io import BytesIO

from mcap_ros2._cdr import CdrReader, CdrWriter

tf2_msg__TFMessage = (
    "0001000001000000cce0d158f08cf9060a000000626173655f6c696e6b00000006000000"
    "7261646172000000ae47e17a14ae0e400000000000000000000000000000000000000000"
    "0000000000000000000000000000000000000000000000000000f03f"
)


def test_serialize():
    buffer = BytesIO()
    writer = CdrWriter(buffer)
    write_example_message(writer)
    assert writer.offset == 100
    assert buffer.getvalue() == bytes.fromhex(tf2_msg__TFMessage)


def test_alignment():
    buffer = BytesIO()
    writer = CdrWriter(buffer)
    assert bytes.hex(buffer.getvalue()) == "00010000"
    writer.write_uint64(1)
    assert bytes.hex(buffer.getvalue()) == "00010000" "0100000000000000"
    writer.write_uint8(2)
    assert bytes.hex(buffer.getvalue()) == "00010000" "0100000000000000" "02"
    writer.write_uint16(3)
    assert (
        bytes.hex(buffer.getvalue()) == "00010000" "0100000000000000" "02" "00" "0300"
    )
    writer.write_uint32(4)
    assert (
        bytes.hex(buffer.getvalue()) == "00010000"
        "0100000000000000"
        "02"
        "00"
        "0300"
        "04000000"
    )
    writer.write_uint8(5)
    writer.write_uint64(6)
    assert (
        bytes.hex(buffer.getvalue()) == "00010000"
        "0100000000000000"
        "02"
        "00"
        "0300"
        "04000000"
        "05"
        "00000000000000"
        "0600000000000000"
    )


def test_roundtrip():
    buffer = BytesIO()
    writer = CdrWriter(buffer)
    writer.write_int8(-1)
    writer.write_uint8(2)
    writer.write_int16(-300)
    writer.write_uint16(400)
    writer.write_int32(-500_000)
    writer.write_uint32(600_000)
    writer.write_int64(-7_000_000_001)
    writer.write_uint64(8_000_000_003)
    writer.write_uint16BE(0x1234)
    writer.write_uint32BE(0x12345678)
    writer.write_uint64BE(0x123456789ABCDEF0)
    writer.write_float32(-9.140000343322754)
    writer.write_float64(1.7976931348623158e100)
    writer.write_string("abc")
    writer.write_uint32(42)
    data = buffer.getvalue()
    assert len(data) == 80

    reader = CdrReader(data)
    assert reader.int8() == -1
    assert reader.uint8() == 2
    assert reader.int16() == -300
    assert reader.uint16() == 400
    assert reader.int32() == -500_000
    assert reader.uint32() == 600_000
    assert reader.int64() == -7_000_000_001
    assert reader.uint64() == 8_000_000_003
    assert reader.uint16BE() == 0x1234
    assert reader.uint32BE() == 0x12345678
    assert reader.uint64BE() == 0x123456789ABCDEF0
    assert reader.float32() == -9.140000343322754
    assert reader.float64() == 1.7976931348623158e100
    assert reader.string() == "abc"
    assert reader.uint32() == 42


def write_example_message(writer: CdrWriter):
    # geometry_msgs/TransformStamped[] transforms
    writer.write_uint32(1)
    # std_msgs/Header header
    writer.write_uint32(1490149580)  # uint32 sec
    writer.write_uint32(117017840)  # uint32 nsec
    writer.write_string("base_link")  # string frame_id
    writer.write_string("radar")  # string child_frame_id
    # geometry_msgs/Transform transform
    writer.write_float64(3.835)  # float64 x
    writer.write_float64(0)  # float64 y
    writer.write_float64(0)  # float64 z
    writer.write_float64(0)  # float64 x
    writer.write_float64(0)  # float64 y
    writer.write_float64(0)  # float64 z
    writer.write_float64(1)  # float64 w



"""the :py:mod:`mcap_ros1.reader` module is deprecated. For similar functionality,
instantiate the :py:class:`mcap.reader.McapReader` with a
:py:class:`mcap_ros1.decoder.DecoderFactory` instance, eg:

>>> from mcap.reader import make_reader
>>> from mcap_ros1.decoder import DecoderFactory
>>> reader = make_reader(open("ros1.mcap", "rb"), decoder_factories=[DecoderFactory()])
>>> for schema_, channel_, message_, ros1_msg in reader.iter_decoded_messages():
>>>     print(ros1_msg)
String(data="hello")
String(data="goodbye")
"""

import warnings
from datetime import datetime
from os import PathLike
from typing import IO, Any, Iterable, Iterator, Optional, Union

from mcap.reader import McapReader, make_reader
from mcap.records import Channel, Message, Schema

from .decoder import DecoderFactory

warnings.warn(__doc__, DeprecationWarning)


def read_ros2_messages(
    source: Union[str, bytes, "PathLike[str]", McapReader, IO[bytes]],
    topics: Optional[Iterable[str]] = None,
    start_time: Optional[Union[int, datetime]] = None,
    end_time: Optional[Union[int, datetime]] = None,
    log_time_order: bool = True,
    reverse: bool = False,
) -> Iterator["McapROS2Message"]:
    """High-level generator that reads ROS2 messages from an MCAP file.

    :param source: some source of MCAP file data. Supply a stream of bytes, an McapReader instance,
        or the path to a valid MCAP file in the filesystem.
    :param topics: an optional list of topics to read from the MCAP file.
    :param start_time: if not None, messages logged before this time will not be included.
    :param end_time: if not None, messages logged at this timestamp or after will not be included.
    :param log_time_order: if True, messages will be yielded in ascending log time order. If
        False, messages will be yielded in the order they appear in the MCAP file.
    :param reverse: if both ``log_time_order`` and ``reverse`` are True, messages will be
        yielded in descending log time order.
    :yields: an McapROS1Message instance for each ROS1 message in the MCAP file.

    .. note::
        this generator assumes the source MCAP conforms to the `ros2` MCAP profile.
        If you need to decode ROS2 messages from a file containing other encodings, use
        the :py:func:`mcap.reader.McapReader.iter_messages()` function to iterate through
        Message records in your MCAP, and decode the ROS2 messages with
        the :py:class:`mcap_ros2.decoder.Decoder` class.
    """
    if start_time is not None and isinstance(start_time, datetime):
        start_time = int(start_time.timestamp() * 1e9)
    if end_time is not None and isinstance(end_time, datetime):
        end_time = int(end_time.timestamp() * 1e9)

    fd = None
    if (
        isinstance(source, PathLike)
        or isinstance(source, str)
        or isinstance(source, bytes)
    ):
        fd = open(source, "rb")
        reader = make_reader(fd, decoder_factories=[DecoderFactory()])
    elif isinstance(source, McapReader):
        reader = source
    else:
        reader = make_reader(source, decoder_factories=[DecoderFactory()])

    try:
        for schema, channel, message, ros2_msg in reader.iter_decoded_messages(
            topics, start_time, end_time, log_time_order, reverse
        ):
            assert schema is not None
            yield McapROS2Message(
                ros_msg=ros2_msg,
                message=message,
                channel=channel,
                schema=schema,
            )
    finally:
        if fd is not None:
            fd.close()


class McapROS2Message:
    """Contains a single ROS2 message and associated metadata.

    :ivar ros_msg: the decoded ROS2 message.
    :ivar sequence_count: the message sequence count if included in the MCAP, or 0 otherwise.
    :ivar log_time_ns: the time this message was logged by the recorder, as a POSIX nanosecond
        timestamp.
    :ivar log_time_ns: the time this message was published, as a POSIX nanosecond
        timestamp.
    :ivar channel: the MCAP Channel record referenced by the Message record
    :ivar schema: the MCAP Schema record referenced by the Channel record
    """

    __slots__ = (
        "ros_msg",
        "sequence_count",
        "log_time_ns",
        "publish_time_ns",
        "channel",
        "schema",
    )

    def __init__(
        self, ros_msg: Any, message: Message, channel: Channel, schema: Schema
    ):
        """
        Construct a new McapROS2Message instance.

        :param ros_msg: the decoded ROS2 message.
        :param message: the MCAP Message record that contains the ROS2 message.
        :param channel: the MCAP Channel record referenced by the Message record.
        """
        self.ros_msg = ros_msg
        self.sequence_count: int = message.sequence
        self.log_time_ns: int = message.log_time
        self.publish_time_ns: int = message.publish_time
        self.channel = channel
        self.schema = schema

    @property
    def log_time(self) -> datetime:
        """Return the timestamp representing when this message was logged by the recorder."""
        return datetime.fromtimestamp(float(self.log_time_ns) / 1e9)

    @property
    def publish_time(self) -> datetime:
        """Return the timestamp representing when this message was published."""
        return datetime.fromtimestamp(float(self.publish_time_ns) / 1e9)



"""ROS2 message definition parsing and message deserialization."""

import os
import re
from io import BytesIO
from types import SimpleNamespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from ._cdr import CdrReader, CdrWriter
from ._vendor.rosidl_adapter.parser import (
    Field,
    MessageSpecification,
    Type,
    parse_message_string,
)

# cSpell:words ftype wstring msgdefs

DecodedMessage = SimpleNamespace
DecoderFunction = Callable[[bytes], DecodedMessage]
EncoderFunction = Callable[[Any], bytes]
PrimitiveValue = Union[bool, int, float, str]
DefaultValue = Union[PrimitiveValue, List[PrimitiveValue]]


def _parseWstring(reader: CdrReader) -> str:
    raise NotImplementedError("wstring parsing is not implemented")


def _parseWstringArray(reader: CdrReader, array_length: int) -> List[str]:
    raise NotImplementedError("wstring[] parsing is not implemented")


def _writeWstring(writer: CdrWriter, value: str):
    raise NotImplementedError("wstring writing is not implemented")


def _writeWstringArray(writer: CdrWriter, value: List[str]):
    raise NotImplementedError("wstring[] writing is not implemented")


FIELD_PARSERS = {
    "bool": CdrReader.boolean,
    "byte": CdrReader.uint8,
    "char": CdrReader.int8,
    "float32": CdrReader.float32,
    "float64": CdrReader.float64,
    "int8": CdrReader.int8,
    "uint8": CdrReader.uint8,
    "int16": CdrReader.int16,
    "uint16": CdrReader.uint16,
    "int32": CdrReader.int32,
    "uint32": CdrReader.uint32,
    "int64": CdrReader.int64,
    "uint64": CdrReader.uint64,
    "string": CdrReader.string,
    "wstring": _parseWstring,
}

ARRAY_PARSERS = {
    "bool": CdrReader.boolean_array,
    "byte": CdrReader.uint8_array,
    "char": CdrReader.int8_array,
    "float32": CdrReader.float32_array,
    "float64": CdrReader.float64_array,
    "int8": CdrReader.int8_array,
    "uint8": CdrReader.uint8_array,
    "int16": CdrReader.int16_array,
    "uint16": CdrReader.uint16_array,
    "int32": CdrReader.int32_array,
    "uint32": CdrReader.uint32_array,
    "int64": CdrReader.int64_array,
    "uint64": CdrReader.uint64_array,
    "string": CdrReader.string_array,
    "wstring": _parseWstringArray,
}

FIELD_WRITERS = {
    "bool": CdrWriter.write_boolean,
    "byte": CdrWriter.write_uint8,
    "char": CdrWriter.write_int8,
    "float32": CdrWriter.write_float32,
    "float64": CdrWriter.write_float64,
    "int8": CdrWriter.write_int8,
    "uint8": CdrWriter.write_uint8,
    "int16": CdrWriter.write_int16,
    "uint16": CdrWriter.write_uint16,
    "int32": CdrWriter.write_int32,
    "uint32": CdrWriter.write_uint32,
    "int64": CdrWriter.write_int64,
    "uint64": CdrWriter.write_uint64,
    "string": CdrWriter.write_string,
    "wstring": _writeWstring,
}

ARRAY_WRITERS = {
    "bool": CdrWriter.write_boolean_array,
    "byte": CdrWriter.write_uint8_array,
    "char": CdrWriter.write_int8_array,
    "float32": CdrWriter.write_float32_array,
    "float64": CdrWriter.write_float64_array,
    "int8": CdrWriter.write_int8_array,
    "uint8": CdrWriter.write_uint8_array,
    "int16": CdrWriter.write_int16_array,
    "uint16": CdrWriter.write_uint16_array,
    "int32": CdrWriter.write_int32_array,
    "uint32": CdrWriter.write_uint32_array,
    "int64": CdrWriter.write_int64_array,
    "uint64": CdrWriter.write_uint64_array,
    "string": CdrWriter.write_string_array,
    "wstring": _writeWstringArray,
}

STRING_TYPES = ("string", "wstring")
FLOAT_TYPES = ("float32", "float64")
INT_TYPES = (
    "byte",
    "char",
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
)

TimeDefinition = MessageSpecification(
    "builtin_interfaces",
    "Time",
    [Field(Type("uint32"), "sec"), Field(Type("uint32"), "nanosec")],
    [],
)


def generate_dynamic(schema_name: str, schema_text: str) -> Dict[str, DecoderFunction]:
    """Convert a ROS2 concatenated message definition into a dictionary of message parsers.

    Modeled after the `generate_dynamic` function in ROS1 `genpy.dynamic`.

    :param schema_name: The name of the schema defined in `schema_text`.
    :param schema_text: The schema text to use for deserializing the message payload.
    :return: A dictionary mapping schema names to message parser.
    """
    msgdefs: Dict[str, MessageSpecification] = {
        "builtin_interfaces/Time": TimeDefinition,
        "builtin_interfaces/Duration": TimeDefinition,
    }
    decoders: Dict[str, DecoderFunction] = {}

    def handle_msgdef(
        cur_schema_name: str, short_name: str, msgdef: MessageSpecification
    ):
        # Add the message definition to the dictionary
        msgdefs[cur_schema_name] = msgdef
        msgdefs[short_name] = msgdef

        # Add the message decoder to the dictionary
        decoder: DecoderFunction = _make_read_message(cur_schema_name, msgdefs)
        decoders[cur_schema_name] = decoder
        decoders[short_name] = decoder

    _for_each_msgdef(schema_name, schema_text, handle_msgdef)
    return decoders


def serialize_dynamic(schema_name: str, schema_text: str) -> Dict[str, EncoderFunction]:
    """Convert a ROS2 concatenated message definition into a dictionary of message encoders.

    :param schema_name: The name of the schema defined in `schema_text`.
    :param schema_text: The schema text to use for serializing message payloads.
    :return: A dictionary mapping schema names to message encoders.
    """
    msgdefs: Dict[str, MessageSpecification] = {
        "builtin_interfaces/Time": TimeDefinition,
        "builtin_interfaces/Duration": TimeDefinition,
    }
    encoders: Dict[str, EncoderFunction] = {}

    def handle_msgdef(
        cur_schema_name: str, short_name: str, msgdef: MessageSpecification
    ):
        # Add the message definition to the dictionary
        msgdefs[cur_schema_name] = msgdef
        msgdefs[short_name] = msgdef

        # Add the message encoder to the dictionary
        encoder: EncoderFunction = _make_encode_message(cur_schema_name, msgdefs)
        encoders[cur_schema_name] = encoder
        encoders[short_name] = encoder

    _for_each_msgdef(schema_name, schema_text, handle_msgdef)
    return encoders


def read_message(
    schema_name: str, msgdefs: Dict[str, MessageSpecification], data: bytes
) -> DecodedMessage:
    """Deserialize a ROS2 message from bytes.

    :param schema_name: The name of the schema to use for deserializing the message payload. This
        key must exist in the `msgdefs` dictionary
    :param msgdefs: A dictionary containing the message definitions for the top-level message and
        any nested messages.
    :param data: The message payload to deserialize.
    :return: The deserialized message.
    """
    msgdef = msgdefs.get(schema_name)
    if msgdef is None:
        raise ValueError(f'Message definition not found for "{schema_name}"')
    reader = CdrReader(data)
    return _read_complex_type(msgdef, msgdefs, reader)


def encode_message(
    schema_name: str, msgdefs: Dict[str, MessageSpecification], ros2_msg: Any
) -> bytes:
    """Serialize a ROS2 message to bytes.

    :param schema_name: The name of the schema to use for deserializing the message payload. This
        key must exist in the `msgdefs` dictionary
    :param msgdefs: A dictionary containing the message definitions for the top-level message and
        any nested messages.
    :param ros2_msg: The message to serialize.
    :return: The serialized message.
    """
    msgdef = msgdefs.get(schema_name)
    if msgdef is None:
        raise ValueError(f'Message definition not found for "{schema_name}"')
    output = BytesIO()
    writer = CdrWriter(output)
    _write_complex_type(msgdef.msg_name, msgdef.fields, msgdefs, ros2_msg, writer)
    return output.getvalue()


def _make_read_message(
    schema_name: str, msgdefs: Dict[str, MessageSpecification]
) -> DecoderFunction:
    return lambda data: read_message(schema_name, msgdefs, data)


def _make_encode_message(
    schema_name: str, msgdefs: Dict[str, MessageSpecification]
) -> EncoderFunction:
    return lambda msg: encode_message(schema_name, msgdefs, msg)


def _read_complex_type(
    msgdef: MessageSpecification,
    msgdefs: Dict[str, MessageSpecification],
    reader: CdrReader,
) -> DecodedMessage:
    Msg = type(
        msgdef.msg_name,
        (SimpleNamespace,),
        {
            "__name__": msgdef.msg_name,
            "__slots__": [field.name for field in msgdef.fields],
            "__repr__": __repr__,
            "__str__": __repr__,
            "__eq__": __eq__,
            "__ne__": __ne__,
            "_type": str(msgdef.base_type),
            "_full_text": str(msgdef),
        },
    )
    msg = Msg()

    if len(msgdef.fields) == 0:
        # In case a message definition definition is empty, ROS 2 adds a
        # `uint8 structure_needs_at_least_one_member` field when converting to IDL,
        # to satisfy the requirement from IDL of not being empty.
        # See also https://design.ros2.org/articles/legacy_interface_definition.html
        reader.uint8()

    for field in msgdef.fields:
        ftype = field.type
        if not ftype.is_primitive_type():
            # Complex type
            nested_definition = msgdefs.get(f"{ftype.pkg_name}/{ftype.type}")
            if nested_definition is None:
                raise ValueError(
                    f'Message definition not found for field "{field.name}" with '
                    'type "{ftype.type}"'
                )

            if ftype.is_array:
                # For dynamic length arrays we need to read a uint32 prefix
                array_length = (
                    ftype.array_size
                    if ftype.is_fixed_size_array() and ftype.array_size is not None
                    else reader.uint32()
                )
                array = [
                    _read_complex_type(
                        nested_definition,
                        msgdefs,
                        reader,
                    )
                    for _ in range(array_length)
                ]
                setattr(msg, field.name, array)
            else:
                value = _read_complex_type(
                    nested_definition,
                    msgdefs,
                    reader,
                )
                setattr(msg, field.name, value)
        else:
            # Primitive type
            if ftype.is_array:
                array_parser_fn = ARRAY_PARSERS.get(ftype.type)
                if array_parser_fn is None:
                    raise NotImplementedError(
                        f"Parsing for type {ftype.type}[] is not implemented"
                    )
                # For dynamic length arrays we need to read a uint32 prefix
                array_length = (
                    ftype.array_size
                    if ftype.is_fixed_size_array() and ftype.array_size is not None
                    else reader.sequence_length()
                )
                value = array_parser_fn(reader, array_length)
                setattr(msg, field.name, value)
            else:
                parser_fn = FIELD_PARSERS.get(ftype.type)
                if parser_fn is None:
                    raise NotImplementedError(
                        f"Parsing for type {ftype.type} is not implemented"
                    )
                value = parser_fn(reader)
                setattr(msg, field.name, value)

    return msg


def _write_complex_type(
    msg_name: str,
    fields: List[Field],
    msgdefs: Dict[str, MessageSpecification],
    ros2_msg: Any,
    writer: CdrWriter,
) -> None:
    if len(fields) == 0:
        # In case a message definition definition is empty, ROS 2 adds a
        # `uint8 structure_needs_at_least_one_member` field when converting to IDL,
        # to satisfy the requirement from IDL of not being empty.
        # See also https://design.ros2.org/articles/legacy_interface_definition.html
        writer.write_uint8(0x00)

    for field in fields:
        ftype = field.type
        if not ftype.is_primitive_type():
            # Complex type
            nested_definition = msgdefs.get(f"{ftype.pkg_name}/{ftype.type}")
            if nested_definition is None:
                raise ValueError(
                    f'Message definition not found for field "{field.name}" with '
                    'type "{ftype.type}"'
                )

            if ftype.is_array:
                array: Union[List[Any], Tuple[Any], Any] = _get_property(
                    ros2_msg, field.name
                )
                if array is None:
                    array = []
                if not isinstance(array, list):
                    raise ValueError(
                        f'Field "{field.name}" is not an array but has array type '
                        f'"{ftype.type}[]"'
                    )

                if ftype.is_fixed_size_array() and ftype.array_size is not None:
                    # Fixed length array, ensure the input array is the correct length
                    while len(array) < ftype.array_size:
                        array.append({})
                    if len(array) > ftype.array_size:
                        array = array[: ftype.array_size]

                    for item in array:
                        _write_complex_type(
                            nested_definition.msg_name,
                            nested_definition.fields,
                            msgdefs,
                            item,
                            writer,
                        )
                else:
                    # Limit the array to the upper bound length, if present
                    if (
                        ftype.is_upper_bound
                        and ftype.array_size is not None
                        and len(array) > ftype.array_size
                    ):
                        array = array[: ftype.array_size]

                    # Dynamic length array, write a uint32 prefix
                    writer.write_uint32(len(array))
                    # Write the array values
                    for item in array:
                        _write_complex_type(
                            nested_definition.msg_name,
                            nested_definition.fields,
                            msgdefs,
                            item,
                            writer,
                        )
            else:
                _write_complex_type(
                    nested_definition.msg_name,
                    nested_definition.fields,
                    msgdefs,
                    _get_property(ros2_msg, field.name) or {},
                    writer,
                )
        else:
            # Primitive type
            if ftype.is_array:
                array: Union[List[Any], Tuple[Any], Any] = _get_property(
                    ros2_msg, field.name
                )
                if array is None:
                    array = []
                if (
                    not isinstance(array, list)
                    and not isinstance(array, tuple)
                    and not isinstance(array, bytes)
                ):
                    raise ValueError(
                        f'Field "{field.name}" is not an array ({type(array)}) but has array type '
                        f'"{ftype.type}[]"'
                    )

                # Special handling for bytes
                if isinstance(array, bytes):
                    byte_array: bytes = array
                    if ftype.type != "uint8" and ftype.type != "byte":
                        raise ValueError(
                            f'Field "{field.name}" has type "uint8[]" but has type "{ftype.type}[]"'
                        )

                    if ftype.is_fixed_size_array() and ftype.array_size is not None:
                        # Fixed length byte array, ensure the input array is the correct length
                        while len(byte_array) < ftype.array_size:
                            byte_array += b"\0"
                        if len(byte_array) > ftype.array_size:
                            byte_array = byte_array[: ftype.array_size]

                        writer.write_bytes(byte_array)
                    else:
                        # Limit the byte array to the upper bound length, if present
                        if (
                            ftype.is_upper_bound
                            and ftype.array_size is not None
                            and len(array) > ftype.array_size
                        ):
                            byte_array = byte_array[: ftype.array_size]

                        # Dynamic length byte array, write a uint32 prefix
                        writer.write_uint32(len(byte_array))
                        # Write the byte array values
                        writer.write_bytes(byte_array)
                else:
                    array_writer_fn = ARRAY_WRITERS.get(ftype.type)
                    if array_writer_fn is None:
                        raise NotImplementedError(
                            f"Writing for type {ftype.type}[] is not implemented"
                        )

                    if ftype.is_fixed_size_array() and ftype.array_size is not None:
                        # Convert tuples to lists
                        list_array = list(array) if isinstance(array, tuple) else array

                        # Fixed length array, ensure the input array is the correct length
                        while len(list_array) < ftype.array_size:
                            list_array.append(None)
                        if len(list_array) > ftype.array_size:
                            list_array = list_array[: ftype.array_size]

                        list_array: List[Any] = _coerce_values(
                            list_array, ftype.type, field.default_value
                        )
                        array_writer_fn(writer, list_array)
                    else:
                        # Limit the array to the upper bound length, if present
                        if (
                            ftype.is_upper_bound
                            and ftype.array_size is not None
                            and len(array) > ftype.array_size
                        ):
                            array = array[: ftype.array_size]

                        array = _coerce_values(array, ftype.type, field.default_value)

                        # Dynamic length array, write a uint32 prefix
                        writer.write_uint32(len(array))
                        # Write the array values
                        array_writer_fn(writer, array)
            else:
                writer_fn = FIELD_WRITERS.get(ftype.type)
                if writer_fn is None:
                    raise NotImplementedError(
                        f"Writing for type {ftype.type} is not implemented"
                    )

                value = _get_property(ros2_msg, field.name)
                value: Any = _coerce_value(value, ftype.type, field.default_value)
                writer_fn(writer, value)


def _for_each_msgdef(
    schema_name: str,
    schema_text: str,
    fn: Callable[[str, str, MessageSpecification], None],
) -> None:
    cur_schema_name = schema_name

    # Remove empty lines
    schema_text = os.linesep.join([s for s in schema_text.splitlines() if s.strip()])

    # Split schema_text by separator lines containing at least 3 = characters
    # (e.g. "===") using a regular expression
    for cur_schema_text in re.split(r"^={3,}$", schema_text, flags=re.MULTILINE):
        cur_schema_text = cur_schema_text.strip()

        # Check for a "MSG: pkg_name/msg_name" line
        match = re.match(r"^MSG:\s+(\S+)$", cur_schema_text, flags=re.MULTILINE)
        if match:
            cur_schema_name = match.group(1)
            # Remove this line from the message definition
            cur_schema_text = re.sub(
                r"^MSG:\s+(\S+)$", "", cur_schema_text, flags=re.MULTILINE
            )

        # Parse the package and message names from the schema name
        # (e.g. "std_msgs/msg/String" -> "std_msgs")
        pkg_name = cur_schema_name.split("/")[0]
        msg_name = cur_schema_name.split("/")[-1]
        short_name = pkg_name + "/" + msg_name
        msgdef = parse_message_string(pkg_name, msg_name, cur_schema_text)

        fn(cur_schema_name, short_name, msgdef)


def _get_property(obj: Any, name: str) -> Any:
    if hasattr(obj, name):
        return getattr(obj, name)
    try:
        return obj[name]
    except (KeyError, TypeError):
        return None


def _coerce_value(
    value: Any, type_name: str, default_value: Optional[DefaultValue]
) -> PrimitiveValue:
    if isinstance(default_value, list):
        raise ValueError("Default value for primitive types cannot be an array")

    if type_name in STRING_TYPES:
        return (
            str(value)
            if value is not None
            else default_value
            if default_value is not None
            else ""
        )
    elif type_name in FLOAT_TYPES:
        return (
            float(value)
            if value is not None
            else default_value
            if default_value is not None
            else 0.0
        )
    elif type_name in INT_TYPES:
        return (
            int(value)
            if value is not None
            else default_value
            if default_value is not None
            else 0
        )
    elif type_name == "bool":
        return (
            bool(value)
            if value is not None
            else default_value
            if default_value is not None
            else False
        )
    else:
        raise NotImplementedError(f'coercion for type "{type_name}" is not implemented')


def _coerce_values(
    values: Union[List[Any], Tuple[Any]],
    type_name: str,
    default_value: Optional[DefaultValue],
) -> List[PrimitiveValue]:
    return [_coerce_value(value, type_name, default_value) for value in values]


def __repr__(self: Any) -> str:
    fields = ", ".join(f"{field}={getattr(self, field)}" for field in self.__slots__)
    return f"{self.__name__}({fields})"


def __eq__(self: Any, other: Any) -> bool:
    if not isinstance(other, type(self)):
        return False

    if (
        not hasattr(self, "__slots__")
        or not hasattr(other, "__slots__")
        or len(self.__slots__) != len(other.__slots__)
    ):
        return False

    for attr in self.__slots__:
        if getattr(self, attr) != getattr(other, attr):
            return False

    return True


def __ne__(self: Any, other: Any) -> bool:
    return not __eq__(self, other)



"""Decoding of CDR (Common Data Representation) data."""

import struct
from enum import IntEnum
from io import BufferedWriter
from typing import Any, BinaryIO, Callable, List, Tuple, Union


class EncapsulationKind(IntEnum):
    """Represents the kind of encapsulation used in a CDR stream."""

    CDR_BE = 0  # Big-endian
    CDR_LE = 1  # Little-endian
    PL_CDR_BE = 2  # Parameter list in big-endian
    PL_CDR_LE = 3  # Parameter list in little-endian


# Avoid the overhead of parsing struct strings every time we need to pack or unpack

unpack_int8 = struct.Struct("b").unpack_from
unpack_uint8 = struct.Struct("B").unpack_from
unpack_int16be = struct.Struct(">h").unpack_from
unpack_int16le = struct.Struct("<h").unpack_from
unpack_uint16be = struct.Struct(">H").unpack_from
unpack_uint16le = struct.Struct("<H").unpack_from
unpack_int32be = struct.Struct(">i").unpack_from
unpack_int32le = struct.Struct("<i").unpack_from
unpack_uint32be = struct.Struct(">I").unpack_from
unpack_uint32le = struct.Struct("<I").unpack_from
unpack_int64be = struct.Struct(">q").unpack_from
unpack_int64le = struct.Struct("<q").unpack_from
unpack_uint64be = struct.Struct(">Q").unpack_from
unpack_uint64le = struct.Struct("<Q").unpack_from
unpack_float32be = struct.Struct(">f").unpack_from
unpack_float32le = struct.Struct("<f").unpack_from
unpack_float64be = struct.Struct(">d").unpack_from
unpack_float64le = struct.Struct("<d").unpack_from

pack_int8 = struct.Struct("b").pack
pack_uint8 = struct.Struct("B").pack
pack_int16be = struct.Struct(">h").pack
pack_int16le = struct.Struct("<h").pack
pack_uint16be = struct.Struct(">H").pack
pack_uint16le = struct.Struct("<H").pack
pack_int32be = struct.Struct(">i").pack
pack_int32le = struct.Struct("<i").pack
pack_uint32be = struct.Struct(">I").pack
pack_uint32le = struct.Struct("<I").pack
pack_int64be = struct.Struct(">q").pack
pack_int64le = struct.Struct("<q").pack
pack_uint64be = struct.Struct(">Q").pack
pack_uint64le = struct.Struct("<Q").pack
pack_float32be = struct.Struct(">f").pack
pack_float32le = struct.Struct("<f").pack
pack_float64be = struct.Struct(">d").pack
pack_float64le = struct.Struct("<d").pack


class CdrReader:
    """Parses values from CDR data."""

    __slots__ = ("data", "offset", "little_endian")

    def __init__(self, data: bytes):
        """Create a CdrReader wrapping a byte array."""
        if len(data) < 4:
            raise ValueError(
                f"Invalid CDR data size {len(data)}, must contain at least a 4-byte header"
            )
        kind = unpack_uint8(data, 1)[0]
        self.data = data
        self.offset = 4
        self.little_endian = kind & 1 == 1

    def kind(self) -> EncapsulationKind:
        """Return the encapsulation kind of the CDR data."""
        return unpack_uint8(self.data, 1)[0]

    def decoded_bytes(self) -> int:
        """Return the number of bytes that have been decoded."""
        return self.offset

    def byte_length(self) -> int:
        """Return the number of bytes in the CDR data."""
        return len(self.data)

    def boolean(self) -> bool:
        """Read an 8-bit value and interpret it as a boolean."""
        return self.uint8() != 0

    def int8(self) -> int:
        """Read a signed 8-bit integer."""
        return self._unpack(unpack_int8, size=1)

    def uint8(self) -> int:
        """Read an unsigned 8-bit integer."""
        return self._unpack(unpack_uint8, size=1)

    def int16(self) -> int:
        """Read a signed 16-bit integer."""
        return self._unpack(
            unpack_int16le if self.little_endian else unpack_int16be, size=2
        )

    def uint16(self) -> int:
        """Read an unsigned 16-bit integer."""
        return self._unpack(
            unpack_uint16le if self.little_endian else unpack_uint16be, size=2
        )

    def int32(self) -> int:
        """Read a signed 32-bit integer."""
        return self._unpack(
            unpack_int32le if self.little_endian else unpack_int32be, size=4
        )

    def uint32(self) -> int:
        """Read an unsigned 32-bit integer."""
        return self._unpack(
            unpack_uint32le if self.little_endian else unpack_uint32be, size=4
        )

    def int64(self) -> int:
        """Read a signed 64-bit integer."""
        return self._unpack(
            unpack_int64le if self.little_endian else unpack_int64be, size=8
        )

    def uint64(self) -> int:
        """Read an unsigned 64-bit integer."""
        return self._unpack(
            unpack_uint64le if self.little_endian else unpack_uint64be, size=8
        )

    def uint16BE(self) -> int:
        """Read an unsigned big-endian 16-bit integer."""
        return self._unpack(unpack_uint16be, size=2)

    def uint32BE(self) -> int:
        """Read an unsigned big-endian 32-bit integer."""
        return self._unpack(unpack_uint32be, size=4)

    def uint64BE(self) -> int:
        """Read an unsigned big-endian 64-bit integer."""
        return self._unpack(unpack_uint64be, size=8)

    def float32(self) -> float:
        """Read a 32-bit floating point number."""
        return self._unpack(
            unpack_float32le if self.little_endian else unpack_float32be, size=4
        )

    def float64(self) -> float:
        """Read a 64-bit floating point number."""
        return self._unpack(
            unpack_float64le if self.little_endian else unpack_float64be, size=8
        )

    def string(self) -> str:
        """Read a string prefixed with its 32-bit length."""
        length = self.uint32()
        if length <= 1:
            # CDR strings are null-terminated, but serializers differ on whether
            # empty strings are length 0 or 1
            self.offset += length
            return ""
        value = self.string_raw(length - 1)
        self.offset += 1  # Skip null terminator
        return value

    def string_raw(self, length: int) -> str:
        """Read a string of the given length."""
        data = self.uint8_array(length)
        value = data.decode("utf-8")
        return value

    def sequence_length(self) -> int:
        """Read a 32-bit unsigned integer."""
        return self.uint32()

    def boolean_array(self, length: int) -> List[bool]:
        """Read an array of booleans of the given length."""
        return [self.uint8() != 0 for _ in range(length)]

    def int8_array(self, length: int) -> List[int]:
        """Read an array of signed 8-bit integers of the given length."""
        return [self.int8() for _ in range(length)]

    def uint8_array(self, length: int) -> bytes:
        """Read a byte array of the given length."""
        data = self.data[self.offset : self.offset + length]
        self.offset += length
        return data

    def int16_array(self, length: int) -> List[int]:
        """Read an array of signed 16-bit integers of the given length."""
        return [self.int16() for _ in range(length)]

    def uint16_array(self, length: int) -> List[int]:
        """Read an array of unsigned 16-bit integers of the given length."""
        return [self.uint16() for _ in range(length)]

    def int32_array(self, length: int) -> List[int]:
        """Read an array of signed 32-bit integers of the given length."""
        return [self.int32() for _ in range(length)]

    def uint32_array(self, length: int) -> List[int]:
        """Read an array of unsigned 32-bit integers of the given length."""
        return [self.uint32() for _ in range(length)]

    def int64_array(self, length: int) -> List[int]:
        """Read an array of signed 64-bit integers of the given length."""
        return [self.int64() for _ in range(length)]

    def uint64_array(self, length: int) -> List[int]:
        """Read an array of unsigned 64-bit integers of the given length."""
        return [self.uint64() for _ in range(length)]

    def float32_array(self, length: int) -> List[float]:
        """Read an array of 32-bit floating point numbers of the given length."""
        return [self.float32() for _ in range(length)]

    def float64_array(self, length: int) -> List[float]:
        """Read an array of 64-bit floating point numbers of the given length."""
        return [self.float64() for _ in range(length)]

    def string_array(self, length: int) -> List[str]:
        """Read an array of strings of the given length."""
        return [self.string() for _ in range(length)]

    def seek(self, relative_offset: int):
        """Seek to a relative offset from the current position."""
        new_offset = self.offset + relative_offset
        if new_offset < 4 or new_offset >= len(self.data):
            raise RuntimeError(
                f"seek({relative_offset}) failed, {new_offset} is outside the data range"
            )
        self.offset = new_offset

    def seek_to(self, offset: int):
        """Seek to an absolute offset."""
        if offset < 4 or offset >= len(self.data):
            raise RuntimeError(
                f"seek_to({offset}) failed, value is outside the data range"
            )
        self.offset = offset

    def _align(self, size: int):
        alignment = (self.offset - 4) % size
        if alignment > 0:
            self.offset += size - alignment

    def _unpack(self, fn: Callable[[bytes, int], Tuple[Any]], size: int):
        if size > 1:
            self._align(size)
        value = fn(self.data, self.offset)[0]
        self.offset += size
        return value


class CdrWriter:
    """Serialize CDR data."""

    __slots__ = ("little_endian", "output", "offset")

    little_endian: bool
    output: Union[BinaryIO, BufferedWriter]
    offset: int

    def __init__(
        self,
        output: Union[BinaryIO, BufferedWriter],
        kind: EncapsulationKind = EncapsulationKind.CDR_LE,
    ):
        """Initialize a CdrWriter wrapping a writable output and write the CDR header."""
        self.little_endian = kind & 1 == 1
        self.output = output
        self.offset = 0
        # Write the CDR header
        self.write_uint16BE(kind)
        self.write_uint16BE(0)

    def write_boolean(self, value: bool):
        """Write a boolean."""
        self.write_uint8(1 if value else 0)

    def write_int8(self, value: int):
        """Write a signed 8-bit integer."""
        self._pack(pack_int8, value, size=1)

    def write_uint8(self, value: int):
        """Write an unsigned 8-bit integer."""
        self._pack(pack_uint8, value, size=1)

    def write_int16(self, value: int):
        """Write a signed 16-bit integer."""
        self._pack(pack_int16le if self.little_endian else pack_int16be, value, size=2)

    def write_uint16(self, value: int):
        """Write an unsigned 16-bit integer."""
        self._pack(
            pack_uint16le if self.little_endian else pack_uint16be, value, size=2
        )

    def write_int32(self, value: int):
        """Write a signed 32-bit integer."""
        self._pack(pack_int32le if self.little_endian else pack_int32be, value, size=4)

    def write_uint32(self, value: int):
        """Write an unsigned 32-bit integer."""
        self._pack(
            pack_uint32le if self.little_endian else pack_uint32be, value, size=4
        )

    def write_int64(self, value: int):
        """Write a signed 64-bit integer."""
        self._pack(pack_int64le if self.little_endian else pack_int64be, value, size=8)

    def write_uint64(self, value: int):
        """Write an unsigned 64-bit integer."""
        self._pack(
            pack_uint64le if self.little_endian else pack_uint64be, value, size=8
        )

    def write_uint16BE(self, value: int):
        """Write an unsigned 16-bit integer in big endian."""
        self._pack(pack_uint16be, value, size=2)

    def write_uint32BE(self, value: int):
        """Write an unsigned 32-bit integer in big endian."""
        self._pack(pack_uint32be, value, size=4)

    def write_uint64BE(self, value: int):
        """Write an unsigned 64-bit integer in big endian."""
        self._pack(pack_uint64be, value, size=8)

    def write_float32(self, value: float):
        """Write a 32-bit floating point number."""
        self._pack(
            pack_float32le if self.little_endian else pack_float32be, value, size=4
        )

    def write_float64(self, value: float):
        """Write a 64-bit floating point number."""
        self._pack(
            pack_float64le if self.little_endian else pack_float64be, value, size=8
        )

    def write_string(self, value: str):
        """Write a string prefixed with its 32-bit length."""
        data = value.encode("utf-8")
        self.write_uint32(len(data) + 1)
        self.write_bytes(data)
        self.output.write(b"\0")  # Null terminator
        self.offset += 1

    def write_bytes(self, value: bytes):
        """Write a byte array."""
        self.output.write(value)
        self.offset += len(value)

    def write_boolean_array(self, values: List[bool]):
        """Write an array of booleans."""
        for value in values:
            self.write_boolean(value)

    def write_int8_array(self, values: List[int]):
        """Write an array of signed 8-bit integers."""
        for value in values:
            self.write_int8(value)

    def write_uint8_array(self, values: List[int]):
        """Write an array of unsigned 8-bit integers."""
        for value in values:
            self.write_uint8(value)

    def write_int16_array(self, values: List[int]):
        """Write an array of signed 16-bit integers."""
        for value in values:
            self.write_int16(value)

    def write_uint16_array(self, values: List[int]):
        """Write an array of unsigned 16-bit integers."""
        for value in values:
            self.write_uint16(value)

    def write_int32_array(self, values: List[int]):
        """Write an array of signed 32-bit integers."""
        for value in values:
            self.write_int32(value)

    def write_uint32_array(self, values: List[int]):
        """Write an array of unsigned 32-bit integers."""
        for value in values:
            self.write_uint32(value)

    def write_int64_array(self, values: List[int]):
        """Write an array of signed 64-bit integers."""
        for value in values:
            self.write_int64(value)

    def write_uint64_array(self, values: List[int]):
        """Write an array of unsigned 64-bit integers."""
        for value in values:
            self.write_uint64(value)

    def write_float32_array(self, values: List[float]):
        """Write an array of 32-bit floating point numbers."""
        for value in values:
            self.write_float32(value)

    def write_float64_array(self, values: List[float]):
        """Write an array of 64-bit floating point numbers."""
        for value in values:
            self.write_float64(value)

    def write_string_array(self, values: List[str]):
        """Write an array of strings, each prefixed with their 32-bit length."""
        for value in values:
            self.write_string(value)

    def _pack(self, fn: Callable[[Any], bytes], value: Any, size: int):
        if size > 1:
            self._align(size)
        self.output.write(fn(value))
        self.offset += size

    def _align(self, size: int):
        # The four byte header is not considered for alignment
        alignment = (self.offset - 4) % size
        padding = size - alignment if alignment > 0 else 0
        if padding > 0:
            self.output.write(b"\0" * padding)
            self.offset += padding



import contextlib
from io import BytesIO
from tempfile import TemporaryFile

from mcap.writer import Writer


class String:
    _type = "std_msgs/msg/String"
    _full_text = "string data"

    def __init__(self, data: str):
        self.data = data

    def serialize(self, buff: BytesIO):
        buff.write(b"\x00\x03")  # CDR header (little-endian, 3)
        buff.write(b"\x00\x00")  # Alignment padding
        buff.write((len(self.data) + 1).to_bytes(4, "little"))  # String length
        buff.write(self.data.encode())  # String data
        buff.write(b"\x00")  # Null terminator


class Empty:
    _type = "std_msgs/msg/Empty"
    _full_text = ""

    def __init__(self):
        pass

    def serialize(self, buff: BytesIO):
        buff.write(b"\x00\x03")  # CDR header (little-endian, 3)
        buff.write(b"\x00\x00")  # Alignment padding
        buff.write(b"\x00")  # uint8 structure_needs_at_least_one_member


@contextlib.contextmanager
def generate_sample_data():
    file = TemporaryFile("w+b")
    writer = Writer(file)
    writer.start(profile="ros2", library="test")
    string_schema_id = writer.register_schema(
        name=String._type, encoding="ros2msg", data=String._full_text.encode()  # type: ignore
    )
    string_channel_id = writer.register_channel(
        topic="/chatter", message_encoding="cdr", schema_id=string_schema_id
    )

    for i in range(10):
        msg = String(data=f"string message {i}")
        buff = BytesIO()
        msg.serialize(buff)  # type: ignore
        writer.add_message(
            channel_id=string_channel_id,
            log_time=i * 1000,
            data=buff.getvalue(),
            publish_time=i * 1000,
        )

    empty_schema_id = writer.register_schema(
        name=Empty._type, encoding="ros2msg", data=Empty._full_text.encode()  # type: ignore
    )
    empty_channel_id = writer.register_channel(
        topic="/empty", message_encoding="cdr", schema_id=empty_schema_id
    )

    for i in range(10):
        msg = Empty()
        buff = BytesIO()
        msg.serialize(buff)  # type: ignore
        writer.add_message(
            channel_id=empty_channel_id,
            log_time=i * 1000,
            data=buff.getvalue(),
            publish_time=i * 1000,
        )

    writer.finish()
    file.seek(0)

    yield file



"""Decoder class for decoding ROS2 messages from MCAP files."""

import warnings
from typing import Any, Callable, Dict, Optional

from mcap.decoder import DecoderFactory as McapDecoderFactory
from mcap.exceptions import McapError
from mcap.records import Message, Schema
from mcap.well_known import MessageEncoding, SchemaEncoding

from ._dynamic import DecodedMessage, DecoderFunction, generate_dynamic


class McapROS2DecodeError(McapError):
    """Raised if a MCAP message record cannot be decoded as a ROS2 message."""

    pass


class DecoderFactory(McapDecoderFactory):
    """Provides functionality to an :py:class:`~mcap.reader.McapReader` to decode CDR-encoded
    messages. Requires valid `ros2msg` schema to decode messages. Schemas written in IDL are not
    currently supported.
    """

    def __init__(self):
        self._decoders: Dict[int, DecoderFunction] = {}

    def decoder_for(
        self, message_encoding: str, schema: Optional[Schema]
    ) -> Optional[Callable[[bytes], DecodedMessage]]:
        if (
            message_encoding != MessageEncoding.CDR
            or schema is None
            or schema.encoding != SchemaEncoding.ROS2
        ):
            return None

        decoder = self._decoders.get(schema.id)
        if decoder is None:
            type_dict = generate_dynamic(  # type: ignore
                schema.name, schema.data.decode()
            )
            if schema.name not in type_dict:
                raise McapROS2DecodeError(f'schema parsing failed for "{schema.name}"')
            decoder = type_dict[schema.name]
            self._decoders[schema.id] = decoder
        return decoder


class Decoder:
    """Decodes ROS 2 messages.

    .. deprecated:: 0.5.0
      Use :py:class:`~mcap_ros2.decoder.DecoderFactory` with :py:class:`~mcap.reader.McapReader`
      instead.
    """

    def __init__(self):
        warnings.warn(
            """The `mcap_ros2.decoder.Decoder` class is deprecated.
For similar functionality, instantiate the `mcap.reader.McapReader` with a
`mcap_ros2.decoder.DecoderFactory` instance.""",
            DeprecationWarning,
        )
        self._decoder_factory = DecoderFactory()

    def decode(self, schema: Schema, message: Message) -> Any:
        decoder = self._decoder_factory.decoder_for(MessageEncoding.CDR, schema)
        assert decoder is not None
        return decoder(message.data)



import time
from io import BufferedWriter
from typing import IO, Any, Dict, Optional, Union

import mcap
from mcap.exceptions import McapError
from mcap.records import Schema
from mcap.well_known import SchemaEncoding
from mcap.writer import CompressionType
from mcap.writer import Writer as McapWriter

from . import __version__
from ._dynamic import EncoderFunction, serialize_dynamic


class McapROS2WriteError(McapError):
    """Raised if a ROS2 message cannot be encoded to CDR with a given schema."""

    pass


def _library_identifier():
    mcap_version = getattr(mcap, "__version__", "<=0.0.10")
    return f"mcap-ros2-support {__version__}; mcap {mcap_version}"


class Writer:
    def __init__(
        self,
        output: Union[str, IO[Any], BufferedWriter],
        chunk_size: int = 1024 * 1024,
        compression: CompressionType = CompressionType.ZSTD,
        enable_crcs: bool = True,
    ):
        self._writer = McapWriter(
            output=output,
            chunk_size=chunk_size,
            compression=compression,
            enable_crcs=enable_crcs,
        )
        self._encoders: Dict[int, EncoderFunction] = {}
        self._channel_ids: Dict[str, int] = {}
        self._writer.start(profile="ros2", library=_library_identifier())
        self._finished = False

    def finish(self):
        """Finishes writing to the MCAP stream. This must be called before the stream is closed."""
        if not self._finished:
            self._writer.finish()
            self._finished = True

    def register_msgdef(self, datatype: str, msgdef_text: str) -> Schema:
        """Write a Schema record for a ROS2 message definition."""
        msgdef_data = msgdef_text.encode()
        schema_id = self._writer.register_schema(
            datatype, SchemaEncoding.ROS2, msgdef_data
        )
        return Schema(
            id=schema_id, name=datatype, encoding=SchemaEncoding.ROS2, data=msgdef_data
        )

    def write_message(
        self,
        topic: str,
        schema: Schema,
        message: Any,
        log_time: Optional[int] = None,
        publish_time: Optional[int] = None,
        sequence: int = 0,
    ):
        """
        Write a ROS2 Message record, automatically registering a channel as needed.

        :param topic: The topic of the message.
        :param message: The message to write.
        :param log_time: The time at which the message was logged as a nanosecond UNIX timestamp.
            Will default to the current time if not specified.
        :param publish_time: The time at which the message was published as a nanosecond UNIX
            timestamp. Will default to ``log_time`` if not specified.
        :param sequence: An optional sequence number.
        """
        encoder = self._encoders.get(schema.id)
        if encoder is None:
            if schema.encoding != SchemaEncoding.ROS2:
                raise McapROS2WriteError(
                    f'can\'t parse schema with encoding "{schema.encoding}"'
                )
            type_dict = serialize_dynamic(  # type: ignore
                schema.name, schema.data.decode()
            )
            # Check if schema.name is in type_dict
            if schema.name not in type_dict:
                raise McapROS2WriteError(f'schema parsing failed for "{schema.name}"')
            encoder = type_dict[schema.name]
            self._encoders[schema.id] = encoder

        if topic not in self._channel_ids:
            channel_id = self._writer.register_channel(
                topic=topic,
                message_encoding="cdr",
                schema_id=schema.id,
            )
            self._channel_ids[topic] = channel_id
        channel_id = self._channel_ids[topic]

        data = encoder(message)

        if log_time is None:
            log_time = time.time_ns()
        if publish_time is None:
            publish_time = log_time
        self._writer.add_message(
            channel_id=channel_id,
            log_time=log_time,
            publish_time=publish_time,
            sequence=sequence,
            data=data,
        )

    def __enter__(self):
        """Context manager support."""
        return self

    def __exit__(self, exc_: Any, exc_type_: Any, tb_: Any):
        """Call finish() on exit."""
        self.finish()



from io import BytesIO

from mcap_ros2.decoder import DecoderFactory
from mcap_ros2.writer import Writer as Ros2Writer

from mcap.reader import make_reader


def read_ros2_messages(stream: BytesIO):
    reader = make_reader(stream, decoder_factories=[DecoderFactory()])
    return reader.iter_decoded_messages()


def test_write_messages():
    output = BytesIO()
    ros_writer = Ros2Writer(output=output)
    schema = ros_writer.register_msgdef("test_msgs/TestData", "string a\nint32 b")
    for i in range(0, 10):
        ros_writer.write_message(
            topic="/test",
            schema=schema,
            message={"a": f"string message {i}", "b": i},
            log_time=i,
            publish_time=i,
            sequence=i,
        )
    ros_writer.finish()

    output.seek(0)
    for index, msg in enumerate(read_ros2_messages(output)):
        assert msg.channel.topic == "/test"
        assert msg.schema.name == "test_msgs/TestData"
        assert msg.decoded_message.a == f"string message {index}"
        assert msg.decoded_message.b == index
        assert msg.message.log_time == index
        assert msg.message.publish_time == index
        assert msg.message.sequence == index


def test_write_std_msgs_empty_messages():
    output = BytesIO()
    ros_writer = Ros2Writer(output=output)
    schema = ros_writer.register_msgdef("std_msgs/msg/Empty", "")
    for i in range(0, 10):
        ros_writer.write_message(
            topic="/test",
            schema=schema,
            message={},
            log_time=i,
            publish_time=i,
            sequence=i,
        )
    ros_writer.finish()

    output.seek(0)
    for index, msg in enumerate(read_ros2_messages(output)):
        assert msg.channel.topic == "/test"
        assert msg.schema.name == "std_msgs/msg/Empty"
        assert msg.message.log_time == index
        assert msg.message.publish_time == index
        assert msg.message.sequence == index
